/*
 Port&Copyright:      Christian Kranz, mail to cnkz at yahoo.com
 Original Copyright:  Hagen Reddmann  mailto HaReddmann at T-Online dot de
 Author:              Hagen Reddmann
 Remarks:             this Copyright must be included
 known Problems:      none
 Version:             0.1. 2005
 Description:         Graphic Library for Siemens S65, M65, CX65 and RX65 Display
*/

.nolist
#include "glcd.inc"
#include "regs.inc"
.list

// grafic library main part
        .section .data
        .global glcd_MemCtrl
        .global glcd_Flags
        .global glcd_Colors
#ifdef USE_CLIPPING
        .global glcd_Clip
#endif

glcd_MemCtrl: .byte   ORIENTATION_DEFAULT
glcd_Flags:   .byte   FLAGS_DEFAULT


glcd_Colors:  .word   SCREEN_COLOR0, SCREEN_COLOR1, SCREEN_COLOR2, SCREEN_COLOR3
  #if COLOR_TABLE_SIZE > 8
              .space  (COLOR_TABLE_SIZE - 8)
  #endif

#ifdef USE_CLIPPING
glcd_Clip:    .byte   SCREEN_LEFT, SCREEN_TOP, SCREEN_RIGHT, SCREEN_BOTTOM
#endif



// initialization procedure of glcd lib
#ifdef USE_AUTOINIT
        .section .init7

glcdInitializeDisplay:
        xcall   glcdDisplayInit
#endif

// glcd library
        .section .text

// SPI Interrupt to reset CS of LCD and deactivate SPI

#if defined (SIG_SPI)
        .global  SIG_SPI
SIG_SPI:
#else
#error "This chip has no SPI interrupts defined"
#endif
        sbi     LCD_PORT, LCD_CS
        cbi     LCD_SPCR, SPE

        reti

;************************************************
//
// void glcdDisplayCommand16(uint16_t cmd);
//
    .global  glcdDisplayCommand16
glcdDisplayCommand16:
    mov     T1,P0H
    rcall   glcdDispCommand
    mov     T1,P0L
    rjmp    glcdDispCommand

//
// void glcdDisplayCommand(uint8_t cmd); 
// send a command to display
    .global  glcdDisplayCommand     // used T1 stack 0
glcdDisplayCommand:
    mov     T1, P0                  // don't touch D0,D1,T0,__tmp_reg__,__zero_reg__
    
    .global glcdDispCommand
glcdDispCommand:
//     set          ; set T flag not needed any more
       sbi     LCD_PORT,LCD_RS        ; set RS line
       cbi     LCD_PORT,LCD_CS        ; select Display
       rjmp    SPI_SEND

//
// void glcdDisplayData(uint8_t data);   
// send data to display
        .global glcdDisplayData         // used T1 stack 0
glcdDisplayData:
    mov     T1, P0          // don't touch P0,D0,D1,T0,__tmp_reg__,__zero_reg__
    .global glcdDispData
glcdDispData:
//    clt       ; clear T flag not needed any more
    .global glcdDispSend
glcdDispSend:   
    cbi     LCD_PORT,LCD_RS     ; clear RS line
    cbi     LCD_PORT,LCD_CS     ; select Display
    ;fall trough

SPI_SEND:   
    sbic    LCD_SPCR,SPE        ;check if SPI ready
    rjmp    SPI_SEND
    sbi     LCD_SPCR,SPE        ; enable SPI
    out     LCD_SPDR,T1         ;load byte in SPI data register

    ret

;************************************************

// void glcdDisplayBuffer(const uint8_t *buffer, uint8_t buffersize, uint8_t flashstored); 
// send a buffer of bytes as data to the display
        .global glcdDisplayBuffer       // used T0,T1,D0,D1,__temp_reg__ r0 if not enhanced AVR  stack 1
glcdDisplayBuffer:
        _movw   ZL, P0
        mov     T0, P1
glcdDispBuffer:
        tst     P2
        brne    glcdDispBuffer2
glcdDispBuffer1:
        ld      T1, Z+
        rcall   glcdDispData
        dec     T0
        brne    glcdDispBuffer1
        ret
glcdDispBuffer2:
        rpm     T1
        rcall   glcdDispData
        dec     T0
        brne    glcdDispBuffer2
        ret


;************************************************

// void glcdWait(uint16_t ms);           
// wait some milliseconds, XTAL must be correct defined

        .global glcdWait                // used T0,T1 stack 0
glcdWait:
        sbiw    P0, 1                                   // 2
        brcc    glcdWait1                               // 2
        ret                                             // 2 -2 = 0 for no branch in both loops
glcdWait1:
        nop
        ldi     T0, lo8(DELAY_ONE_MILLISECOND -4)       // 1
        ldi     T1, hi8(DELAY_ONE_MILLISECOND -4)       // 1
glcdWait2:
        sbiw    T0, 1                                   // 2
        brne    glcdWait2                               // 2 = 4, = 3 for last loop
        rjmp    glcdWait                                // 2 = 8


;************************************************


; write table with r30/r31 pointer to table
; r22 number of bytes to read
wrtab:
    lpm     T1,Z+           ; load byte from table
    rcall   glcdDispCommand
    dec     r22
    brne    wrtab
    ret

//
// void glcdDisplayInit(void);
        .global glcdDisplayInit         // used D0,D1,T0,T1,__tmp_reg__,__zero_reg__ stack 4
glcdDisplayInit:

    push    r30
    push    r31


    sbi     LCD_PORT, LCD_CS        ;deselect Display
    sbi     LCD_DDR,  LCD_CS        ;

    
    cbi     LCD_PORT, LCD_RESET     ; display reset
    sbi     LCD_DDR,  LCD_RESET     ; set reset as output
;   sbi     LCD_DDR,  LCD_MISO      ; MISO is always input in SPI
    sbi     LCD_DDR,  LCD_MOSI      ; MOSI must be set as output for SPI
    sbi             LCD_PORT, LCD_CS        ; deselect display 
    sbi     LCD_DDR,  LCD_CS        ; SS must be output for Master mode to work


    cbi     LCD_PORT, LCD_SCK       ; set SCK lo
    sbi     LCD_DDR,  LCD_SCK       ; set SCK as output
    sbi     LCD_PORT, LCD_MISO      ; enable pull up of MISO to avoid floating input

    sbi     LCD_PORT, LCD_RS        ; start with RS=high
    sbi     LCD_DDR, LCD_RS         ; RS signal is output

    ; setup serial data interface 
    ; select clock phase positive going in middle of data
    ; master mode
    ; enable SPI
    ; speed is CPUclock/2

    ldi     r24,1               ; double speed bit
    out     LCD_SPSR,r24

    ldi     r24,(1<<MSTR)  |(1<<SPIE); | (1<<SPR0); ; Master Mode,  Enable SPI, fCPU/16 clock
    out     LCD_SPCR,r24

    // make a longer reset...
    ldi     P0L,1   ; wait x ms
    ldi     P0H,0
    rcall   glcdWait

    sbi     LCD_PORT, LCD_RESET ; release reset
    in      r24,LCD_SPSR        ; clear status    
        sei
    ; ports and SPI are now configured
    
    
    ldi     r30,lo8(INIT1)  ;set pointer to table
    ldi     r31,hi8(INIT1)       
    ldi     r22,24          ;number of bytes to be transmitted
    rcall   wrtab
        

    ; wait for settling power chip, 31.2ms
    ; 22.8ms ... 31.2ms
    ; 38ms is too much, 15.6ms is too low
    ldi     P0L,6   ; wait x ms
    ldi     P0H,0
    rcall   glcdWait

    
    ldi     r30,lo8(INIT2)  ;set pointer to table
    ldi     r31,hi8(INIT2)       
    ldi     r22,42          ;number of bytes to be transmitted
    rcall   wrtab

;   write initial screen command (i think not necessary...)
    cbi LCD_PORT,LCD_CS ;select Display

    ldi T1,0xEF
    rcall   glcdDispCommand
    ldi T1,0x90
    rcall   glcdDispCommand
    ldi T1,0x00
    rcall   glcdDispCommand
    ldi T1,0x00
    rcall   glcdDispCommand

        rcall   lcd_clrscr

    pop     r31
    pop     r30
    ret


;************************************************
;clear LCD
;        r24
;void lcd_clrscr(void)
    .global lcd_clrscr
lcd_clrscr:
    push    r30
    push    r31

    ; send window write start sequence
    ldi T1,0xEF
    rcall   glcdDispCommand
    ldi T1,0x90
    rcall   glcdDispCommand
    ldi T1,0x05
    rcall   glcdDispCommand
    ldi T1,0x00
    rcall   glcdDispCommand
    ldi T1,0x06
    rcall   glcdDispCommand
    ldi T1,0x00
    rcall   glcdDispCommand
    ldi T1,0x07
    rcall   glcdDispCommand
    ldi T1,0x00
    rcall   glcdDispCommand


    ; now 132*176*2 bytes are to be transmitted
    ; mask  is 0xF800 for red
    ;          0x07E0 for green
    ;          0x001F for blue
    ldi     r31, 0xB5
    ldi r30, 0x40
    ldi T1, 0x00
LCD_CLEAR_1:
    ldi T1, 0x00
    rcall   glcdDispData
    sbiw    r30,1
    brne    LCD_CLEAR_1

    pop r31
    pop r30
    ret



; Init tables for S65-Display
INIT1:  ; 24bytes
.byte 0xFD, 0xFD, 0xFD, 0xFD
.byte 0xEF, 0x00, 0xEE, 0x04, 0x1B, 0x04, 0xFE, 0xFE, 0xFE, 0xFE, 0xEF, 0x90, 0x4A, 0x04
.byte 0x7F, 0x3F, 0xEE, 0x04, 0x43, 0x06 

INIT2:  ; 42bytes
.byte 0xEF, 0x90,  0x09, 0x83,  0x08, 0x00,  0x0B, 0xAF,  0x0A, 0x00,  0x05, 0x00,  0x06, 0x00
.byte 0x07, 0x00,  0xEF, 0x00,  0xEE, 0x0C,  0xEF, 0x90,  0x00, 0x80,  0xEF, 0xB0,  0x49, 0x02 
.byte 0xEF, 0x00,  0x7F, 0x01,  0xE1, 0x81,  0xE2, 0x02,  0xE2, 0x76,  0xE1, 0x83,  0x80, 0x01




.end














